#include "MeMegaPi.h"
#include <Servo.h>
MePort limitSwitch(PORT_7);

Servo svs[2] = {Servo(),Servo()};
MeStepperOnBoard steppers[3] = {MeStepperOnBoard(PORT_1),MeStepperOnBoard(PORT_2),MeStepperOnBoard(PORT_3)}; 
float angles[3] = {0,0,0};

struct Vector3
{
  double x;
  double y; 
  double z;
};

float lastPositions[3] = {0,0,0};
float initAngleY = 36;
float initAngleZ = 54;
const double RADS = PI / 180.0;
const double DEGS = 180.0 / PI;
const int STEPS = 16;
const int GEARS = 7;
const double L1 = 144.0; 
const double L2 = 176.0; 
String buffer = "";
bool endMoving = true;
Vector3 vectorToAngles(float x,float y,float z);
void setSpeed(float speed);

int currentSpeed = 2000;
int pin1;
int pin2;
void setup() {
  Serial.begin(115200);
  for(int i=0;i<3;i++){
    steppers[i].setSpeed(currentSpeed);
    steppers[i].setMaxSpeed(currentSpeed);
    steppers[i].setAcceleration(currentSpeed);
    steppers[i].setMicroStep(STEPS);
    steppers[i].enableOutputs();
  }
  pin1 = limitSwitch.pin1();
  pin2 = limitSwitch.pin2();
  pinMode(pin1,INPUT_PULLUP);
  pinMode(pin2,INPUT_PULLUP);
  svs[0].attach(A8);
  svs[1].attach(A9);
  reset();
}
//G1 X0 Y160 Z0
void reset(){
  int limits[2] = {0,0};
  for(int i=0;i<3;i++){
    steppers[i].setSpeed(1000);
  }
  while(true){
    if(digitalRead(pin1)==LOW){
      steppers[1].step();
      delayMicroseconds(500);
    }else{
      limits[0] = 1;
      if(digitalRead(pin2)==LOW){
        steppers[2].step();
        delayMicroseconds(500);
      }else{
        limits[1] = 1;
      }
    }
    if(limits[0]==1&&limits[1]==1){
      break;
    }
  }
  steppers[1].setCurrentPosition(-initAngleY*STEPS*GEARS/1.8);
  steppers[2].setCurrentPosition(-initAngleZ*STEPS*GEARS/1.8);
  lastPositions[0] = 0;
  lastPositions[1] = 227;
  lastPositions[2] = 0;
  moveToPoint(lastPositions[0],lastPositions[1],lastPositions[2]);
}
void loop() {
  if (Serial.available()) {
    char c = Serial.read();
    if (c == '\n') {
      parseBuffer();
    } else {
      buffer += c;
    }
  }
  
  long isMoving = 0;
    for(int i=0;i<3;i++){
      isMoving += abs(steppers[i].distanceToGo());
      steppers[i].run();
    }
    if(isMoving>0){
      endMoving = true;
    }else{
    if(endMoving){
      endMoving = false;
      Serial.println("ok");
    }
    }
}
void parseBuffer() {
  
  buffer =" "+buffer+" ";
  buffer.toLowerCase();
  int count = 0;
  int startIndex = 0;
  int endIndex = 0;
  int len = buffer.length();
  if (len < 1) {
    return;
  }
  String tmp;
  String values[6];
  bool gCodeEnable = false;
  bool mCodeEnable = false;
  while (true) {
    startIndex = buffer.indexOf(" ", endIndex);
    endIndex = buffer.indexOf(" ", startIndex + 1);
    tmp = buffer.substring(startIndex + 1, endIndex);
    
    if(tmp.indexOf("x",0)>-1){
      values[1] = tmp.substring(1, tmp.length());
      lastPositions[0] = stringToFloat(values[1]);
    }else if(tmp.indexOf("y",0)>-1){
      values[2] = tmp.substring(1, tmp.length());
      lastPositions[1] = stringToFloat(values[2]);
    }else if(tmp.indexOf("z",0)>-1){
      values[3] = tmp.substring(1, tmp.length());
      lastPositions[2] = stringToFloat(values[3]);
    }else if(tmp.indexOf("f",0)>-1){
      values[4] = tmp.substring(1, tmp.length());
      setSpeed(stringToFloat(values[4]));
    }else if(tmp.indexOf("g",0)>-1){
      values[0] = tmp.substring(1, tmp.length());
      gCodeEnable = true;
    }else if(tmp.indexOf("m",0)>-1){
      values[0] = tmp.substring(1, tmp.length());
      mCodeEnable = true;
    }else if(tmp.indexOf("s",0)>-1){
      values[1] = tmp.substring(1, tmp.length());
    }else if(tmp.indexOf("p",0)>-1){
      values[2] = tmp.substring(1, tmp.length());
    }
    count++;
    if (endIndex == len - 1) break;
  }
  int code = atoi(values[0].c_str());
  if(gCodeEnable){
    switch (code) {
      case 1:
      moveToPoint(lastPositions[0],lastPositions[1],lastPositions[2]);
      break;
      case 28:
      reset();
      break;
    }
  }else if(mCodeEnable){
  switch (code) {
    case 1:
    runServo(stringToFloat(values[1]),stringToFloat(values[2]));
    break;
    case 3:
    runLaser(stringToFloat(values[2]));
    break;
  }
  Serial.println("ok");
  }
  buffer = "";
}
void runServo(int index,int angle){
  svs[index].write(angle);
}
void runLaser(int power){
  pinMode(10,OUTPUT);
  analogWrite(10,power);
}
void moveToPoint(float x,float y,float z){
  Vector3 t = vectorToAngles(x,y,z);
  steppers[0].moveTo(t.x*STEPS*GEARS/1.8);
  steppers[1].moveTo(t.y*STEPS*GEARS/1.8);
  steppers[2].moveTo((t.z-(t.y+initAngleY))*STEPS*GEARS/1.8);
}
void setSpeed(float speed){
  if(currentSpeed != speed){
    currentSpeed = speed;
    for(int i=0;i<3;i++){
        steppers[i].setSpeed(speed);
        steppers[i].setMaxSpeed(speed);
        steppers[i].setAcceleration(speed);
    }
  }
}
Vector3 vectorToAngles(float x,float y,float z){
  long pow1 = L1*L1;
  long pow2 = L2*L2;
  long powRH = x*x+y*y;
  double RH = sqrt(powRH);
  long powRV = z*z+RH*RH;
  double RV = sqrt(powRV);
  Vector3 a;
  a.x = atan2(x,y)*DEGS;//*STEPS*GEARS/1.8;
  a.y = ((atan2(-z,RH)-acos((powRV+pow1-pow2)/(2*RV*L1)))*DEGS+90.0)-90;//*STEPS*GEARS/1.8;
  a.z = (acos((pow1+pow2-powRV)/(2*L1*L2))-PI/2)*DEGS-90;//*STEPS*GEARS/1.8;
  
  return a;
}
float stringToFloat(String s){
  float f = atof(s.c_str());
  return f;
}